"""
clickext.decorators

Argument and option decorators for clickext commands.
"""

from __future__ import annotations

import json
import logging
from pathlib import Path
import typing as t

import click
import tomli
import yaml

try:
    from yaml import CSafeLoader as SafeLoader
except ImportError:  # pragma: no cover
    from yaml import SafeLoader

from .core import ClickextCommand
from .log import init_logging

if t.TYPE_CHECKING:
    from .log import Styles


_AnyCallable: t.TypeAlias = t.Callable[..., t.Any]
FC = t.TypeVar("FC", bound=_AnyCallable | ClickextCommand)


def config_option(
    file: Path | str,
    *param_decls: str,
    processor: t.Optional[_AnyCallable] = None,
    require_config: bool = False,
    **kwargs: t.Any,
) -> t.Callable[[FC], FC]:
    """Adds a configuration file option.

    Provides a method to load, parse, and optionally prepare data from a configuration file. The result is saved to
    `ctx.obj` and can be accessed with `@click.pass_context`, `@click.pass_obj`, or by registering a
    `click.make_pass_decorator` for the prepared data object type. This option is non-eager so the configuration is not
    loaded when the program will not run (for example, when "--help" or "--version" is passed).

    Configuration files must be a JSON, TOML, or YAML file. The file extension determines the file format:

        - JSON:

            - .json

        - TOML:

            - .toml

        - YAML:

            - .yaml
            - .yml

    The config option itself is always optional, however setting `require_config` to `True` will prevent the program
    starting if a configuration file is not present. If a configuration file is missing, is not required, and a
    `processor` is specified, the processor will be passed the `None` value for program-specific handling, otherwise
    `ctx.obj` will be set to `None`.

    By default this option uses the `click.Path` parameter type and returns a `pathlib.Path` object to the loader
    callback. It will also accept a `str` via the `click.STRING` parameter type if the program cannot use `click.Path`
    for some reason.

    :param file: The default configuration file location.
    :param param_decls: One or more option names. (Default: "--config / -c").
    :param processor: An optional callable that receives the parsed config file and prepares it per the program's
    specifications. This callable must accept a `None` value and return the prepared data or object.
    :param require_config: Whether a configuration file is required to start the program.

    :raises click.ClickException: When the configuration file 1) is required and doesn't exist, 2) cannot be read,
    3) cannot be parsed, or 4) is an unknown format.
    """

    def callback(
        ctx: click.Context, param: click.Parameter, value: Path | str  # pylint: disable=unused-argument
    ) -> None:
        if isinstance(value, str):
            value = Path(value)

        raw_text = None

        if value.is_file():
            try:
                raw_text = value.read_text(encoding="utf8")
            except OSError as exc:
                raise click.ClickException("Failed to read configuration file") from exc
        elif require_config:
            raise click.ClickException("Configuration file not found")

        if raw_text:
            try:
                match value.suffix:
                    case ".json":
                        config = json.loads(raw_text)
                    case ".toml":
                        config = tomli.loads(raw_text)
                    case ".yaml" | ".yml":
                        config = yaml.load(raw_text, Loader=SafeLoader)
                    case _:
                        raise click.ClickException(f'Unknown configuration file format "{value.suffix}"')
            except (json.JSONDecodeError, tomli.TOMLDecodeError, yaml.YAMLError) as exc:
                raise click.ClickException("Failed to parse configuration file") from exc
        else:
            config = None

        if processor:
            config = processor(config)

        ctx.obj = config

    if not param_decls:
        param_decls = ("--config", "-c")

    kwargs.setdefault("expose_value", False)
    kwargs.setdefault("help", "The configuration file to use.")
    kwargs.setdefault("type", click.Path(path_type=Path))
    kwargs["default"] = str(file)
    kwargs["is_eager"] = False
    kwargs["callback"] = callback
    kwargs["required"] = False

    return click.option(*param_decls, **kwargs)


def verbose_option(
    logger: logging.Logger,
    *param_decls: str,
    root_handlers: t.Optional[list[logging.Handler]] = None,
    prefix_styles: t.Optional[dict[int, t.Optional[Styles]]] = None,
    **kwargs: t.Any,
) -> t.Callable[[FC], FC]:
    """Adds a verbose option.

    A flag to switch between standard output and verbose output. The `--verbose` flag should be passed before any other
    eager options to ensure the desired verbosity level is set before the other options are evaluated. This option
    initializes the logging environment so it is not necessary to call `log.init_logging` when this option is used.

    :param logger: The program logger. Required to initialize logging. See `clickext.log.init_logging`.
    :param param_decls: One or more option names. (Default: "--verbose / -v").
    :param root_handlers: Additional handlers to attach to the root logger. See `clickext.log.init_logging`.
    :param prefix_styles: Log level prefix display styles. See: `clickext.log.ConsoleFormatter` for format.
    :param kwargs: Extra arguments passed to `click.option`.
    """

    def callback(ctx: click.Context, param: click.Parameter, value: bool) -> None:  # pylint: disable=unused-argument
        root_logger = logging.getLogger()
        root_logger.setLevel(logging.DEBUG if value else logging.INFO)
        logger.setLevel(root_logger.level)
        logging.raiseExceptions = root_logger.level == logging.DEBUG

    init_logging(logger, root_handlers=root_handlers, prefix_styles=prefix_styles)

    if not param_decls:
        param_decls = ("--verbose", "-v")

    kwargs.setdefault("metavar", "LVL")
    kwargs.setdefault("expose_value", False)
    kwargs.setdefault("help", "Increase verbosity.")
    kwargs["is_flag"] = True
    kwargs["flag_value"] = True
    kwargs["default"] = False
    kwargs["is_eager"] = True
    kwargs["callback"] = callback

    return click.option(*param_decls, **kwargs)


def verbosity_option(
    logger: logging.Logger,
    *param_decls: str,
    root_handlers: t.Optional[list[logging.Handler]] = None,
    prefix_styles: t.Optional[dict[int, t.Optional[Styles]]] = None,
    **kwargs: t.Any,
) -> t.Callable[[FC], FC]:
    """Adds a configurable verbosity option.

    The `--verbosity` flag should be passed before any other eager options to ensure the desired verbosity level is set
    before the other options are evaluated. This option initializes the logging environment so it is not necessary to
    call `log.init_logging` when this option is used. Available verbosity levels are (from least to most verbose):

        - "QUIET"
        - "CRITICAL"
        - "ERROR"
        - "WARNING"
        - "INFO" (DEFAULT)
        - "DEBUG"

    Levels are case-insensitive.

    :param logger: The program logger. Required to initialize logging. See `clickext.log.init_logging`.
    :param param_decls: One or more option names. (Default: "--verbosity / -v").
    :param root_handlers: Additional handlers to attach to the root logger. See `clickext.log.init_logging`.
    :param prefix_styles: Log level prefix display styles. See: `clickext.log.ConsoleFormatter` for format.
    :param kwargs: Extra arguments passed to `click.option`.
    """

    def callback(ctx: click.Context, param: click.Parameter, value: str) -> None:  # pylint: disable=unused-argument
        root_logger = logging.getLogger()
        root_logger.setLevel(getattr(logging, value.upper()))
        logger.setLevel(root_logger.level)
        logging.raiseExceptions = logger.getEffectiveLevel() == logging.DEBUG

    init_logging(logger, level=kwargs.get("default", "INFO"), root_handlers=root_handlers, prefix_styles=prefix_styles)

    if not param_decls:
        param_decls = ("--verbosity", "-v")

    kwargs.setdefault("default", "INFO")
    kwargs.setdefault("metavar", "LVL")
    kwargs.setdefault("expose_value", False)
    kwargs.setdefault("help", "Specify verbosity level.")
    kwargs["is_eager"] = True
    kwargs["type"] = click.Choice(["QUIET", "CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"], case_sensitive=False)
    kwargs["callback"] = callback

    return click.option(*param_decls, **kwargs)



"""
clickext

The clickext public API
"""

from .core import ClickextCommand as ClickextCommand
from .core import ClickextGroup as ClickextGroup
from .decorators import config_option as config_option
from .decorators import verbose_option as verbose_option
from .decorators import verbosity_option as verbosity_option
from .log import init_logging as init_logging



"""
clickext.core

Extended functionality for the click library.
"""

import errno
import typing as t

import click


__all__ = ["ClickextCommand", "ClickextGroup"]


class ClickextCommand(click.Command):
    """A clickext command.

    Clickext commands support aliases and marking options as mutually exclusive.

    Aliases have no effect unless the command is a subcommand in a `ClickextGroup`.

    Mutually exclusive options are validated before invoking the command. Validation fails when all of the mutually
    exclusive options are passed as arguments and one or more options has a value other than its default.

    :param aliases: Alternate names that should invoke this command.
    :param mx_opts: Groups of options that are mutually exclusive. Each item is a `tuple` of `click.Option` names that
    cannot be used together, e.g, `[("foo", "bar")]`.
    :param catch_exceptions: Whether exceptions that occur during command invocation should be caught and re-raised as
    `click.ClickExceptions`.
    :param attrs: Extra arguments passed to `click.Command`.
    """

    def __init__(
        self,
        name: t.Optional[str] = None,
        aliases: t.Optional[list[str]] = None,
        mx_opts: t.Optional[list[tuple[str]]] = None,
        catch_exceptions: bool = True,
        **attrs: t.Any,
    ):
        super().__init__(name, **attrs)

        self._catch_exceptions = catch_exceptions
        self.aliases = sorted(aliases or [])
        self.mx_opts = mx_opts or []
        self.global_opts: dict[str, click.Option] = {}

    def invoke(self, ctx: click.Context) -> t.Any:
        """Given a context, this invokes the command.

        Catches most exceptions that occur during a command invocation and re-raises them as a `click.ClickException`.
        This provides a simple way to handle all program errors in the command line context while allowing the code to
        raise relevant exceptions in other contexts without using `try... except` blocks around code in every command to
        prevent errors bubbling up. This behavior can be disabled by setting `catch_exceptions=False` in the command
        definition.

        The logging facility, if enabled by calling `clickext.init_logging`, will continue to format and handle
        exception output according to the logging configuration whether this is enabled or not.

        The following exceptions are not caught:

            - `EOFError`
            - `KeyboardInterrupt`
            - `OSError` (when `OSError.errno == errno.EPIPE`)
            - `click.Abort`
            - `click.ClickException`
            - `click.exceptions.Exit`

        :param ctx: The current `click.Context` object.

        :raises click.clickException: When an unlisted exception occurs during invocation.
        """
        try:
            return super().invoke(ctx)
        except (EOFError, KeyboardInterrupt, OSError, click.Abort, click.ClickException, click.exceptions.Exit) as exc:
            if self._catch_exceptions and isinstance(exc, OSError) and exc.errno != errno.EPIPE:
                raise click.ClickException(str(exc)) from exc
            raise
        except Exception as exc:  # pylint: disable=broad-except
            if self._catch_exceptions:
                raise click.ClickException(str(exc)) from exc
            raise

    def parse_args(self, ctx: click.Context, args: t.List[str]) -> t.List[str]:
        """Parse arguments and update the context.

        Mutually exclusive options are validated after parsing because the resolved values are required to determine
        which were options passed and which are using default values.

        :param ctx: The current `click.Context` object.
        :param args: A list of arguments passed to the program.
        """
        args = super().parse_args(ctx, args)
        self.validate_mutually_exclusive_options(ctx)
        return args

    def validate_mutually_exclusive_options(self, ctx: click.Context) -> None:
        """Ensure mutually exclusive options have not been passed as arguments.

        Validation must be done after the arguments are parsed so only the options relevant to the command are
        considered and the resolved values can be compared to the default value for each option.

        :param ctx: The current `click.Context` object.

        :raises click.UsageError: When mutually exclusive options have been passed.
        """
        for mx_opts in self.mx_opts:
            passed: list[str] = []

            for param in self.get_params(ctx):
                if param.name in mx_opts and param.name in ctx.params and ctx.params[param.name] != param.default:
                    passed.append(param.opts[0])

                    if len(passed) == len(mx_opts):
                        raise click.UsageError(f"Mutually exclusive options: {' '.join(passed)}", ctx)

    def list_global_options(self) -> list[click.Option]:
        """Get a list of global options registered with the command."""
        return list(self.global_opts.values())

    def format_help(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        """Writes the help into the formatter if it exists.

        :param ctx: The current `click.Context` object.
        :param formatter: The help output formatter.
        """
        super().format_help(ctx, formatter)
        self.format_aliases(ctx, formatter)

    def format_options(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        """Add options to the program help display.

        Options are sorted alphabetically by the first CLI option string. If the command is part of a `ClickextGroup`
        global group options are included in the option display.

        :param ctx: The current `click.Context` object.
        :param formatter: The help output formatter.
        """
        params = self.get_params(ctx)

        if ctx.parent and not isinstance(self, ClickextGroup):
            params.extend(self.list_global_options())

        params.sort(key=lambda x: x.opts[0])

        opts = []

        for param in params:
            record = param.get_help_record(ctx)
            if record is not None:
                opts.append(record)

        if opts:  # pragma: no branch
            with formatter.section("Options"):
                formatter.write_dl(opts)

    def format_aliases(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        """Add aliases to the program help display when command is a subcommand.

        :param ctx: The current `click.Context` object.
        :param formatter: The help output formatter.
        """
        if ctx.parent:
            aliases = []

            for alias in self.aliases:
                aliases.append((alias, ""))

            if aliases:
                with formatter.section("Aliases"):
                    formatter.write_dl(aliases)


class ClickextGroup(ClickextCommand, click.Group):
    """A clickext command group.

    Clickext command groups require `ClickextCommand` command instances. Groups support aliases, global options, shared
    subcommand options, and mutually exclusive options.

    Global options are options defined at the group level that may be passed as arguments to the group command
    and all subcommands in the group. Typically these options are used to change configuration or execution globally
    (e.g., to set verbosity level). Global options are extracted and prepended to the arguments prior to argument
    parsing; they are not passed to subcommands and must not be part of a subcommand function signature.

    Global option names cannot be the same as a subcommand name, subcommand option name, or share long/short option
    strings with a subcommand option. Additionally, non-flag global options should not accept values that begin with "-"
    or values identical to a subcommand name. Global options can be mutually exclusive with other group-level options,
    but not with subcommand options.

    Shared parameters are parameters defined at the group level that are attached to all subcommands in the group. These
    parameters cannot be passed to the group itself. Shared parameter names cannot be the same as a subcommand parameter
    name or share long/short option strings with a subcommand option. Shared parameters cannot be mutually exclusive
    with global options, but can be mutually exclusive with non-shared subcommand parameters.

    :param global_opts: `click.Option` names that can be passed to the group and all subcommands in the group.
    :param shared_params: `click.Parameter` names that can be passed to all subcommands in the group, but not the group
    itself.
    :param attrs: Extra arguments passed to `ClickextCommand` and `click.Group`.
    """

    def __init__(
        self,
        name: t.Optional[str] = None,
        global_opts: t.Optional[list[str]] = None,
        shared_params: t.Optional[list[str]] = None,
        **attrs: t.Any,
    ):
        super().__init__(name, **attrs)

        self.global_opts: dict[str, click.Option] = self.init_global_options(global_opts)
        self.shared_params: dict[str, click.Parameter] = self.init_shared_params(shared_params)

    def init_global_options(self, names: t.Optional[list[str]] = None) -> dict[str, click.Option]:
        """Find and validate global options for the group.

        :param names: A list of `click.Option` names to make global.

        :raises ValueError: When an option with the given name does not exist.
        :raises TypeError: When a parameter is not a `click.Option`.
        """
        names = names or []
        options: dict[str, click.Option] = {}

        for name in names:
            param = next((p for p in self.params if p.name == name), None)

            if not param:
                raise ValueError(f"Unknown global option {name}")

            if not isinstance(param, click.Option):
                raise TypeError(f"Invalid global option {name}; global options must be a 'click.Option'")

            options[name] = param

        return options

    def init_shared_params(self, names: t.Optional[list[str]] = None) -> dict[str, click.Parameter]:
        """Find, extract, and validate shared subcommand parameters from the group parameters.

        :param names: A list of parameter names to share with all subcommands.

        :raises ValueError: When a parameter with the given name does not exist.
        """
        names = names or []
        params: dict[str, click.Parameter] = {}

        for name in names:
            param = next((p for p in self.params if p.name == name), None)

            if not param:
                raise ValueError(f"Unknown shared parameter {name}")

            params[name] = param

        # subcommand parameters cannot be passed to the group
        self.params = [p for p in self.params if p.name not in names]

        return params

    def make_parser(self, ctx: click.Context, globals_only: bool = False) -> click.OptionParser:
        """Creates the underlying option parser for this command.

        :param ctx: The current `click.Context` object.
        :param globals_only: Whether the parser should be restricted to global options.
        """
        parser = click.OptionParser(ctx)
        global_options = self.list_global_options()

        for param in self.get_params(ctx):
            if not globals_only or (globals_only and param in global_options):
                param.add_to_parser(parser, ctx)

        return parser

    def parse_args(self, ctx: click.Context, args: list[str]) -> list[str]:
        """Parse arguments and update the context.

        When global options are defined parsing will occur in two passes. The first pass parses global options
        exclusively. Global options (and their values, if applicable) are extracted from the original args and the args
        list is rebuilt with the global options at the beginning. The reconstructed args are then passed to the parent
        to be processed by click normally.

        :param ctx: The current `click.Context` object.
        :param args: A list of arguments passed to the program.
        """
        args = self.parse_global_args(ctx, args)
        return super().parse_args(ctx, args)

    def parse_global_args(self, ctx: click.Context, args: list[str]) -> list[str]:
        """Parse global arguments and rebuild the args list.

        :param ctx: The current `click.Context` object.
        :param args: A list of arguments passed to the program.
        """
        if not self.global_opts:
            return args

        original_ctx_settings = {
            "allow_extra_args": None,
            "allow_interspersed_args": None,
            "ignore_unknown_options": None,
        }

        # Save the original context settings to restore after parsing. Global parsing requires `ctx.allow_extra_args`,
        # `ctx.allow_interspersed_args`, and `ctx.ignore_unknown_options` to be `True`.
        for setting in original_ctx_settings:
            original_ctx_settings[setting] = getattr(ctx, setting)
            setattr(ctx, setting, True)

        parser = self.make_parser(ctx, globals_only=True)

        try:
            param_order: list[click.Option]  # globals can only be `click.Option`
            opts, args, param_order = parser.parse_args(args)  # type: ignore
        finally:
            for setting, value in original_ctx_settings.items():
                setattr(ctx, setting, value)

        global_args = []

        for param in param_order:
            global_args.append(param.opts[0])

            if param.is_flag or not param.name:
                continue

            value = opts[param.name]

            if value and value != param.flag_value:  # pragma: no branch
                if isinstance(value, tuple):
                    global_args.extend(value)
                else:
                    global_args.append(value)

        return global_args + args

    def add_command(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, cmd: ClickextCommand, name: t.Optional[str] = None
    ) -> None:
        """Register a command with this group.

        The command must be a `ClickextCommand`. Global options are stored separately on the command so they are
        available in the program help display.

        :param cmd: A `ClickextCommand` to add to the group.
        :param name: The command name.
        :raises TypeError: When a command that is not a `ClickextCommand` is added to the group.
        """
        if not isinstance(cmd, ClickextCommand):
            raise TypeError("Only 'ClickextCommand's can be registered with a 'ClickextGroup'")

        name = name or cmd.name

        if name is not None:  # pragma: no branch
            self.validate_command(cmd, name)
            cmd.global_opts = self.global_opts
            cmd.params.extend(self.list_shared_parameters())

        super().add_command(cmd, name)

    def get_command(self, ctx: click.Context, cmd_name: str) -> t.Optional[ClickextCommand]:
        """Get a command by name or alias.

        :param ctx: The current `click.Context` object.
        :param name: The command name.
        """
        for name, cmd in self.commands.items():
            if cmd_name == name or cmd_name in getattr(cmd, "aliases", []):
                cmd_name = name
                break

        return super().get_command(ctx, cmd_name)  # pyright: ignore[reportReturnType]

    def validate_command(self, cmd: ClickextCommand, name: str) -> None:
        """Validate global options and shared parameters for a command.

        :param cmd: A `ClickextCommand` to validate.
        :param name: The command name.

        :raises ValueError: When the command name is the same as a global option name, or a command parameter conflicts
        with a global option, shared parameter name, or option string.
        """
        if name in self.global_opts:
            raise ValueError(f"Subcommand {name} conflicts with a global option name")

        for param in cmd.params:
            if param.name in self.global_opts:
                raise ValueError(f"Subcommand option {param.name} conflicts with a global option name")

            if param.name in self.shared_params:
                raise ValueError(
                    f"Subcommand {param.param_type_name} {param.name} conflicts with a shared parameter name"
                )

            for opt in param.opts:
                if any(opt in gopt.opts for gopt in self.list_global_options()):
                    raise ValueError(f"Subcommand option string {opt} conflicts with a global option string")

                if any(opt in sopt.opts for sopt in self.list_shared_parameters()):
                    raise ValueError(f"Subcommand option string {opt} conflicts with a shared option string")

    def list_shared_parameters(self) -> list[click.Parameter]:
        """Get a list of parameters shared with all subcommands."""
        return list(self.shared_params.values())

    def format_commands(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        """Add subcommands to the program help display.

        :param ctx: The current `click.Context` object.
        :param formatter: The help output formatter.
        """
        commands: list[tuple[str, click.Command]] = []

        for subcommand in self.list_commands(ctx):
            cmd = self.get_command(ctx, subcommand)

            if cmd is None or cmd.hidden:
                continue

            aliases = getattr(cmd, "aliases", [])

            if aliases:
                subcommand = f"{subcommand} ({','.join(aliases)})"

            commands.append((subcommand, cmd))

        # allow for 3 times the default spacing
        if commands:
            limit = formatter.width - 6 - max(len(cmd[0]) for cmd in commands)
            rows = []

            for subcommand, cmd in commands:
                cmd_help = cmd.get_short_help_str(limit)
                rows.append((subcommand, cmd_help))

            if rows:  # pragma: no branch
                with formatter.section("Commands"):
                    formatter.write_dl(rows)

    def format_options(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        """Add options and commands to the program help display.

        Options are sorted alphabetically by the first CLI option string.

        :param ctx: The current `click.Context` object.
        :param formatter: The help output formatter.
        """
        super().format_options(ctx, formatter)
        self.format_commands(ctx, formatter)



"""
clickext.log

Logging and console output handling for clickext programs.
"""

import logging
import textwrap
import typing as t

import click

from .exceptions import patch_exceptions


QUIET_LEVEL_NAME = "QUIET"
QUIET_LEVEL_NUM = 1000


class Styles(t.TypedDict, total=False):
    """Style types for `click.style`"""

    fg: t.Optional[int | t.Tuple[int, int, int] | str]
    bg: t.Optional[int | t.Tuple[int, int, int] | str]
    bold: t.Optional[bool]
    dim: t.Optional[bool]
    underline: t.Optional[bool]
    overline: t.Optional[bool]
    italic: t.Optional[bool]
    blink: t.Optional[bool]
    reverse: t.Optional[bool]
    strikethrough: t.Optional[bool]
    reset: bool


class ConsoleFormatter(logging.Formatter):
    """Format log messages for the console.

     By default, "INFO" level messages are passed through as-is. All other levels are formatted with a level name
     prefix: "{level:} {msg}".

    :param prefix_styles: A mapping of log level numbers to `click.style` parameters used to style the prefix in the
    formatted message. Style parameters are merged with the defaults unless styles is set to `None` in which case
    messages for that level will not be prefixed with the level name. Unknown levels are silently ignored. Example:

    ```
    prefix_styles = {
        logging.CRITICAL: {"fg": "purple"}, // fg=purple (overrides default)
        logging.DEBUG: {"bg": "green"}, // fg=blue, bg=green (merged with default)
        logging.ERROR: None // (will not prefix or style ERROR level messages)
    }
    ```
    """

    _default_styles: dict[int, Styles] = {
        logging.CRITICAL: {"fg": "red"},
        logging.DEBUG: {"fg": "blue"},
        logging.ERROR: {"fg": "red"},
        logging.INFO: {},
        logging.WARNING: {"fg": "yellow"},
    }

    def __init__(self, *, prefix_styles: t.Optional[dict[int, t.Optional[Styles]]] = None):
        super().__init__()
        self.prefix_styles: dict[int, t.Optional[Styles]] = {}

        if prefix_styles is None:
            prefix_styles = {}

        for level, style in self._default_styles.items():
            style_override: t.Optional[Styles] = {}

            if level in prefix_styles:
                style_override = prefix_styles[level]

            self.prefix_styles[level] = {} if style_override is None else {**style, **style_override}

    def format(self, record: logging.LogRecord) -> str:
        record.message = record.getMessage().strip()

        prefix_style = self.prefix_styles[record.levelno]

        if prefix_style:
            prefix = click.style(f"{record.levelname.title()}:", **prefix_style)
            record.message = f"{prefix} {record.message}"
            record.message = textwrap.indent(
                record.message, str(" " * (len(record.levelname) + 2)), lambda x: not x.startswith(prefix)
            )

        if record.exc_info:
            record.exc_text = self.formatException(record.exc_info)
            record.message = f"{record.message}\n{record.exc_text}"

        return record.message


class ColorFormatter(ConsoleFormatter):
    """For backwards compatibility; use `ConsoleFormatter` instead."""


class ConsoleHandler(logging.Handler):
    """Send log messages to the console.

    Writes to stderr if the record level is `logging.WARNING` or greater, otherwise to stdout.
    """

    def emit(self, record: logging.LogRecord) -> None:
        try:
            msg = self.format(record)
            click.echo(msg, err=record.levelno >= logging.WARNING)
        except Exception:  # pylint: disable=broad-except
            self.handleError(record)


def init_logging(
    logger: logging.Logger,
    level: int | str = logging.INFO,
    root_handlers: t.Optional[list[logging.Handler]] = None,
    prefix_styles: t.Optional[dict[int, t.Optional[Styles]]] = None,
) -> None:
    """Initialize program logging.

    Configures the root logger to handle all log records and warnings issued with `warnings.warn` with `ConsoleHandler`
    and the `ConsoleFormatter`. Existing handlers are removed from the root logger.

    `click.ClickException`, `click.UsageError`, and their children are patched so their message output is sent to a
    program logger instead of printing to the console directly so the output is formatted consistently and can
    conditionally supressed based on the log level.

    An additional log level is added during initialization and assigned to `logging.QUIET`. This level can be used to
    suppress all log record output.

    All informational messages must be sent to a logger since direct calls to `print`, `click.echo`, etc. cannot be
    captured or formatted. Generic messages can be logged at the `INFO` level which, by default, outputs the message
    as-received with no additional formatting. Non-informational messages that should always be displayed (e.g, the
    calculated result of the program) should continue to be sent with `click.echo`.

    :param logger: The program logger. Patched click exception messages will be sent to this logger in order to preserve
    the source of the message. This will usually be the logger corresponding to the program's entrypoint module.
    :param level: The log level to print (default: `logging.INFO`).
    :param root_handlers: Additional handlers to attach to the root logger.
    :param prefix_styles: Log level prefix display styles. See: `ConsoleFormatter` for format.
    """
    if not hasattr(logging, QUIET_LEVEL_NAME):
        logging.addLevelName(QUIET_LEVEL_NUM, QUIET_LEVEL_NAME)
        setattr(logging, QUIET_LEVEL_NAME, QUIET_LEVEL_NUM)

    if isinstance(level, str):
        level = getattr(logging, level.upper(), logging.INFO)

    if root_handlers is None:
        root_handlers = []

    root_logger = logging.getLogger()
    root_logger.handlers.clear()

    console_formatter = ConsoleFormatter(prefix_styles=prefix_styles)
    console_handler = ConsoleHandler()
    console_handler.setFormatter(console_formatter)

    for handler in [console_handler, *root_handlers]:
        root_logger.addHandler(handler)

    root_logger.setLevel(level)

    logging.captureWarnings(True)
    logging.raiseExceptions = level == logging.DEBUG

    logger.setLevel(level)
    logger.propagate = True

    patch_exceptions(logger)



"""
clickext.exceptions

Logger-aware exception handling.
"""

from __future__ import annotations

import logging
import sys
import typing as t

import click

if t.TYPE_CHECKING:
    from types import TracebackType


def patch_exceptions(logger: logging.Logger) -> None:
    """Sends click and other uncaught exceptions to a logger.

    Patches `click.ClickException`, `click.UsageError`, their children, and `sys.excepthook` to override the default
    behavior of printing a message directly to the console. Instead, messages will be printed consistent with the
    current log level providing greater control of the program output. Exception information, including a traceback when
    available, will be printed if the log level is `logging.DEBUG`.

    This function is called automatically by `clickext.init_logging` when the program logging is initialized. It should
    not be called manually.

    :param logger: The program logger. See `clickext.log.init_logging`.
    """

    def excepthook(exc_type: type[BaseException], exc_value: BaseException, exc_traceback: TracebackType) -> None:
        exc_info = (exc_type, exc_value, exc_traceback)

        if issubclass(exc_type, KeyboardInterrupt):
            sys.__excepthook__(*exc_info)
            return

        logger.critical(str(exc_value), exc_info=exc_info if logger.level == logging.DEBUG else None)

    click.ClickException.logger = logger  # pyright: ignore[reportAttributeAccessIssue]
    click.ClickException.show = _click_exception_patch
    click.UsageError.show = _click_usage_error_patch
    sys.excepthook = excepthook


def _click_exception_patch(self: click.ClickException, file: t.Optional[t.IO] = None) -> None:
    """Patch for `click.ClickException.show` that sends output a logger."""
    file = click.get_text_stream("stderr") if file is None else file
    exc_info = (
        self
        if self.logger.getEffectiveLevel() == logging.DEBUG  # pyright: ignore[reportAttributeAccessIssue]
        else None
    )
    self.logger.error(self.format_message(), exc_info=exc_info)  # pyright: ignore[reportAttributeAccessIssue]


def _click_usage_error_patch(self: click.UsageError, file: t.Optional[t.IO] = None) -> None:
    """Patch for `click.UsageError.show` that sends output a logger."""
    file = click.get_text_stream("stderr") if file is None else file
    exc_info = (
        self
        if self.logger.getEffectiveLevel() == logging.DEBUG  # pyright: ignore[reportAttributeAccessIssue]
        else None
    )
    hint = ""

    if self.ctx is not None:
        hint = ""

        if self.ctx.command.get_help_option(self.ctx) is not None:
            hint = f"Try '{self.ctx.command_path} {self.ctx.help_option_names[0]}' for help.\n"

        click.echo(f"{self.ctx.get_usage()}\n{hint}", file=file, color=None)

    self.logger.error(self.format_message(), exc_info=exc_info)  # pyright: ignore[reportAttributeAccessIssue]
